/*
    c99tree - parse tree output layer, output format: abstract C
    Copyright (C) 2010-2015  Tibor Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "c99tree-cc.h"
#include "tree_print.h"

typedef struct c_print_s {
	c99tree_cc_t *ctx;
	c99c_node_t *next_token;
	int space_pending;
	int auto_ws;
} c_print_t;

static const char *WHITESPACE = " ";
static const char *NEWLINE = "\n";

/* token tree sync */
#define pend_space(pctx) pctx->space_pending = 1

static int tokenstr_is_ws(char *str)
{
	if ((str[0] == '/') && (str[1] == '*'))
		return 1;
	if ((str[0] == '/') && (str[1] == '/'))
		return 1;

	for(;*str != '\0'; str++) {
		if (!isspace(*str))
			return 0;
	}
	return 1;
}

static void print_tt_upto(c_print_t *pctx, const c99c_node_t *gram_node, const char *alt_str, int inclusive)
{
	c99c_node_t *ok = NULL, *n, *tok_node = gram_node->loc.twin_parent;

	if (pctx->auto_ws)  /* TODO: temporary hack until real abstract print */
		printf("%s", (alt_str == NULL ? " " : alt_str));

	if (tok_node == NULL) {
		if ((pctx->auto_ws) && (inclusive)) /* TODO: temporary hack until real abstract print */
			printf("%s", gram_node->name == NULL ? "" : gram_node->name);
		else if ((inclusive) && (alt_str != NULL))
			printf("%s", alt_str);
		return;
	}

	pctx->space_pending = 0;

	if (inclusive) {
		ok = tok_node->loc.twin_parent;
		tok_node = tok_node->next;
	}

	for(n = pctx->next_token; (n != tok_node) && (n != NULL); n = n->next) {
		if ((n->loc.twin_parent != NULL) && (n->loc.twin_parent != ok)) {
			fflush(stdout);
			fprintf(stderr, "broken tt 1 (ok=%lu tok_node=%lu got %lu)\n", ok == NULL ? 0 : ok->uid, tok_node == NULL ? -1 : tok_node->uid, n->loc.twin_parent->uid);
			abort();
		}
		printf("%s", n->name);
		pctx->next_token = n->next;
		if ((n->name != NULL) && (alt_str != NULL)) {
			if ((alt_str == WHITESPACE) || (alt_str == NEWLINE)) {
				if (strpbrk(n->name, " \t\r\n") != NULL)
					alt_str = NULL;
			}
			else if (strcmp(n->name, alt_str) == 0)
				alt_str = NULL;
		}
	}
	if (alt_str != NULL)
		printf("%s", alt_str);
	pctx->next_token = n;
}

/* Print ws tokens until (including) the next one matches bump. Alternatively
   stop when the next token has a grammar twin (and then print bump) */
static void print_tt_ws_upto(c_print_t *pctx, const char *bump)
{
	c99c_node_t *n;


	for(n = pctx->next_token; n != NULL; n = n->next) {
		int eq = strcmp(n->name, bump) == 0;
		if ((n->loc.twin_parent != NULL) || (!eq && !isspace(*n->name))) { /* reached the first no-no token */
			printf("%s", bump);
			return;
		}
		printf("%s", n->name);
		pctx->next_token = n->next;
		if (eq) /* printed a token that happened to match */
			return;
	}
	pctx->next_token = NULL;
}

static void print_tt_token_(c_print_t *pctx, const c99c_node_t *gram_node, const char *str, int insert, int ws_target_only)
{
	c99c_node_t *n;

	if (gram_node != NULL) {
		print_tt_upto(pctx, gram_node, str, 1);
		return;
	}

	for(n = pctx->next_token; n != NULL; n = n->next) {
		int can_comp = ((n->name != NULL) && (str != NULL));
		int eq = 0;
		if (can_comp)
			eq = (strcmp(n->name, str) == 0);

		if ((n->loc.twin_parent != NULL) && (!eq))
			break;

		if (ws_target_only) {
			if (!eq && !tokenstr_is_ws(n->name))
				break;
		}

		pctx->next_token = n->next;
		printf("%s", n->name);
		if (can_comp) {
			if (eq) {
				return;
			}
			if ((str == WHITESPACE) || (str == NEWLINE))
				if (strpbrk(n->name, " \t\r\n") != NULL)
					return;
		}
	}
	if ((insert) && (str != NULL))
		printf("%s", str);
}

static void print_tt_token(c_print_t *pctx, const c99c_node_t *gram_node, const char *str)
{
	print_tt_token_(pctx, gram_node, str, 1, 0);
}

static void print_tt_eat(c_print_t *pctx, const c99c_node_t *gram_node, const char *str)
{
	print_tt_token_(pctx, gram_node, str, 0, 1);
}


static void print_tt_id(c_print_t *pctx, const c99c_node_t *gram_node)
{
	char *name = gram_node->name && strcmp(gram_node->name, "<anonymous>") != 0 ? gram_node->name : "";
	c99c_node_t *tok_node = gram_node->loc.twin_parent;

	print_tt_upto(pctx, gram_node, NULL, 0);

	printf("%s", name);

	if (tok_node == NULL)
		return;

	if ((tok_node->loc.twin_parent != NULL) && (tok_node->loc.twin_parent != gram_node)) {
		fprintf(stderr, "broken tt 2\n");
		abort();
	}

	pctx->next_token = tok_node->next;
}

static void print_tt_rest(c_print_t *pctx)
{
	c99c_node_t *n;

	pctx->space_pending = 0;

	for(n = pctx->next_token; n != NULL; n = n->next) {
		if (n->loc.twin_parent != NULL) {
			fflush(stdout);
			fprintf(stderr, "broken tt 3\n");
			abort();
		}
		printf("%s", n->name);
	}
	pctx->next_token = NULL;
}


static void c_print_nodes(c_print_t *pctx, const c99c_node_t *node, const char *sep, int prec);
static void c_print_node(c_print_t *pctx, const c99c_node_t *node, int prec);
static void c_print_attr(c_print_t *pctx, const c99c_node_t *n, int prec);
static void c_print_attrs(c_print_t *pctx, const c99c_node_t *parent_node, int allow_pre, int prec);
static void c_print_pre_attrs(c_print_t *pctx, const c99c_node_t *parent_node, int prec);
static void c_print_id(c_print_t *pctx, const c99c_node_t *node);

/* empty string == prec level increase */
static char *opname[] = {
/* 0*/	"", /* block expr */
/* 1*/	"", /* exprlist */
/* 2*/	"", /* arg */
/*  */	/* ellipsis and assignment on the same level, since they don't appear together */
/* 3*/	"...", "=", "|=", "^=", "&=", ">>=", "<<=", "+=", "-=", "*=", "/=", "%=", "",
/* 4*/	"?", "",
/* 5*/	"||", "",
/* 6*/	"&&", "",
/* 7*/	"|", "",
/* 8*/	"^", "",
/* 9*/	"&", "",
/*10*/	"==", "!=", "",
/*11*/	"<", ">", "<=", ">=", "",
/*12*/	"<<", ">>", "",
/*13*/	"+", "-", "",
/*14*/	"*", "/", "%", "",
/*15*/	"cast", "",
/*16*/	"&", "*", "+", "-", "~", "!", "pre++", "pre--", "sizeof", "", /* unary */
/*17*/	"[]", ".", "->", "post++", "post--", /* (), cast with initlist */
/*  */	0
};

/* if child prec < parent prec then child needs () */
/* parent prec can be forged in some cases, some child never has () */
static int getprec(c_print_t *pctx, const c99c_node_t *n) {
	char *name;
	char *mode;
	char **p;
	int prec;

	switch (n->type) {
	case T_EXPRBRACKET: return 0;
	case T_EXPRLIST: return 1;
	case T_INITLIST:
	case T_ARG: return 2; /* >exprlist */
	case T_PTR: return 16;
	case T_INDEX:
	case T_ARGLIST: /* only for decl */
	case T_CALL: return 17; /* post */
	case T_TYPE: return 15; /* <sizeof */
	case T_INIT: return 3; /* assignment */
	case T_CONDITION:
		if (n->name && strcmp(n->name, "condition") != 0)
			return 0; /* in for(cond1;;cond3) no () around assignment */
		return 4; /* >assignment */
	case T_OP: break;
	default: return 0;
	}
	name = c99c_strclone(n->name);
	mode = strchr(name, ' ');
	if (mode == NULL)
		goto badop;
	*mode = '\0';
	mode++;
	switch (*mode) {
	case 'u': /* unary */
		switch (*name) {
		case '&': case '*': case '+': case '-':
			free(name);
			return 16;
		}
	case 't': /* ternary */
	case 'b': /* binary */
		break;
	default: badop:
		fprintf(stderr, "ERROR: invalid operator (needs unary/binary/trenary)\n");
		exit(1);
	}
	for (p = opname, prec = 0; *p; p++) {
		if (**p == 0) {
			prec++;
			continue;
		}
		if (strcmp(*p, name) == 0)
			break;
	}
	if (*p == 0) {
		fprintf(stderr, "ERROR: unknown operator: %s\n", n->name);
		exit(1);
	}
	free(name);
	return prec;
}

static void declid(c_print_t *pctx, const c99c_node_t *n, const c99c_node_t *id_node, int pprec) {
	int prec;

	if (n == NULL) {
		print_tt_id(pctx, id_node);
		return;
	}
	prec = getprec(pctx, n);
	switch (n->type) {
	case T_TYPE:
		declid(pctx, n->children, id_node, pprec);
		break;
	case T_PTR:
		if (prec < pprec)
			print_tt_eat(pctx, NULL, "(");

		print_tt_token(pctx, n, "*");

		declid(pctx, n->next, id_node, prec);
		if (prec < pprec)
			print_tt_eat(pctx, NULL, ")");
		break;
	case T_INDEX:
		declid(pctx, n->next, id_node, prec);

		print_tt_token(pctx, n, "[");

		c_print_nodes(pctx, n->children, NULL, 0);
		/* vla f(int n, int a[*]), but not abstract_n */
		if (n->name != NULL)
			if (strcmp(n->name, "*") == 0)
				print_tt_token(pctx, n, NULL);

		print_tt_token(pctx, NULL, "]");

		break;
	case T_QUALIFIER:
		/* >16 == not pointer */
		if (16 < pprec) {
			fprintf(stderr, "ERROR: unexpected qualifier: %s:%s\n", c99c_node_type_names[n->type], n->name);
			exit(1);
		}
		c_print_node(pctx, n, prec);
		declid(pctx, n->next, id_node, prec);
		break;
	case T_ARGLIST:
		if ((n->name != NULL) && (strcmp(n->name, "K&R/1") == 0)) {
			if (n->next == NULL) {
				print_tt_token(pctx, id_node, NULL);
				print_tt_token(pctx, NULL, "(");
				c_print_nodes(pctx, n->children, ",", 0);
				print_tt_token(pctx, NULL, ")");
			}
			else {
				declid(pctx, n->next->next, id_node, prec); /* must be NULL */
				c_print_node(pctx, n, 0); /* K&R/1 */
				c_print_node(pctx, n->next, 0); /* K&R/2 */
			}
		} else {
			declid(pctx, n->next, id_node, prec);
			c_print_node(pctx, n, 0);
		}
		break;
	case T_INIT: /* fix it later */
		declid(pctx, n->next, id_node, 0);
		break;
	case T_BITS:
		declid(pctx, n->next, id_node, 0);
		c_print_node(pctx, n, 0);
		break;
	case T_ASM:
		print_tt_token(pctx, NULL, "asm");
		print_tt_token(pctx, NULL, "(");
		c_print_node(pctx, n->children, prec);
		print_tt_token(pctx, NULL, ")");
		declid(pctx, n->next, id_node, 0);
		break;
	case T_ATTRLIST:
		/* print non-post attributes as part of id's type to preserve the right order */
		if (!attrlist_is_post(n))
			c_print_attr(pctx, n, 0);
		declid(pctx, n->next, id_node, 0);
		break;
	default:
		fprintf(stderr, "ERROR: invalid token in declaration: %s:%s\n", c99c_node_type_names[n->type], n->name);
		exit(1);
	}
}

static void c_print_id(c_print_t *pctx, const c99c_node_t *node)
{

	if (node->children) { /* decl */
		c99c_node_t *p;
		declid(pctx, node->children, node, 0);
		/* ASM comes before post-attribs */
		for (p = node->children; p; p = p->next)
			if (p->type == T_ASM)
				c_print_node(pctx, p, 0);

		/* post-attribs */
		c_print_attrs(pctx, node, 0, 0);

		/* INIT goes at the end */
		for (p = node->children; p; p = p->next)
			if (p->type == T_INIT)
				c_print_node(pctx, p, 0);
	}
	else {
#warning TODO: this else is not needed, the above code can handle this case fine
		print_tt_id(pctx, node);
		c_print_attrs(pctx, node, 0, 0);
	}
}

static void c_print_op(c_print_t *pctx, const c99c_node_t *node, int prec)
{
	char *mode, *name;

	name = c99c_strclone(node->name);
	mode = strchr(name, ' ');
	if (mode == NULL) {
		fprintf(stderr, "ERROR: invalid operator (needs unary/binary/trenary)\n");
		exit(1);
	}
	*mode = '\0';
	mode++;
	if (strcmp(mode, "unary") == 0) {
		if (strcmp(name, "post++") == 0) {
			c_print_node(pctx, node->children, prec);
			print_tt_token(pctx, node, NULL);
		} else if (strcmp(name, "post--") == 0) {
			c_print_node(pctx, node->children, prec);
			print_tt_token(pctx, node, NULL);
		} else if (strcmp(name, "pre++") == 0) {
			print_tt_token(pctx, node, "++");
			c_print_node(pctx, node->children, prec);
		} else if (strcmp(name, "pre--") == 0) {
			print_tt_token(pctx, node, "--");
			c_print_node(pctx, node->children, prec);
		} else {
			print_tt_token(pctx, node, NULL);
			/* sizeof i */
			c_print_node(pctx, node->children, prec);
		}
	}
	else if (strcmp(mode, "binary") == 0) {
		if (*name == '[') {
			c_print_node(pctx, node->children, prec);
			print_tt_token(pctx, node, "[");
			c_print_node(pctx, node->children->next, 0);
			print_tt_token(pctx, NULL, "]");
		} else if (strcmp(name, "cast") == 0) {
			c_print_node(pctx, node->children, prec+1); /* type needs () */
			c_print_node(pctx, node->children->next, prec);
		} else {
			char opname[32], *si, *so;
			int n;
			c_print_node(pctx, node->children, prec);
			/* extra space to avoid a-(--b) and 3...4 (gcc extension) */
			for(si = node->name, so = opname, n = 0; (*si != ' ') && (*si != '\0') && n < sizeof(opname)-1; si++,so++)
				*so = *si;
			*so = '\0';
			print_tt_token(pctx, node, opname);
			c_print_node(pctx, node->children->next, prec+1);
		}
	}
	else if (strcmp(mode, "trenary") == 0) {
		c_print_node(pctx, node->children, prec+1);
		print_tt_token(pctx, node, "?");
		c_print_node(pctx, node->children->next, 0);
		print_tt_token(pctx, NULL, ":");
		c_print_node(pctx, node->children->next->next, prec);
		}
	else {
		fprintf(stderr, "ERROR: invalid operator (has to be unary/binary/trenart instead of '%s')\n", mode);
		exit(1);
	}
}

static void c_print_statement(c_print_t *pctx, const c99c_node_t *node)
{
	if (node->name == NULL) {
		print_tt_token(pctx, node, ";");
		return;
	}
	switch(*node->name) {
		case 'e': /* expr */
			c_print_nodes(pctx, node->children, NULL, 1); /* block is () */
			print_tt_token(pctx, NULL, ";");
			break;
		case 'f': /* for */
			print_tt_token(pctx, node, NULL);
			print_tt_token(pctx, NULL, "(");
			c_print_node(pctx, node->children, 0);
			if (node->children->children == NULL || node->children->children->type != T_DECL)
				print_tt_token(pctx, NULL, ";");
			c_print_node(pctx, node->children->next, 0);
			print_tt_token(pctx, NULL, ";");
			c_print_node(pctx, node->children->next->next, 0);
			print_tt_token(pctx, NULL, ")");
			if (node->children->next->next->next == NULL)
				print_tt_token(pctx, NULL, ";");
			else
				c_print_node(pctx, node->children->next->next->next, 0);
			break;
		case 'w': /* while */
			print_tt_token(pctx, node, NULL);
			print_tt_token(pctx, NULL, "(");
			c_print_node(pctx, node->children, 0);
			print_tt_token(pctx, NULL, ")");
			c_print_node(pctx, node->children->next, 0);
			break;
		case 'd': /* do */
			print_tt_token(pctx, node, NULL);
			if (node->children->next != NULL) {
				c_print_node(pctx, node->children, 0);
				print_tt_token(pctx, NULL, "while");
				print_tt_token(pctx, NULL, "(");
				c_print_node(pctx, node->children->next->children, 0);
				print_tt_token(pctx, NULL, ")");
			}
			else {
/*				printf("; while(");
				c_print_node(pctx, node->children->children, 0);
				printf(");");*/
				abort();
			}
			break;
		case 'i': /* if */
			print_tt_token(pctx, node, NULL);
			print_tt_token(pctx, NULL, "(");
			c_print_node(pctx, node->children, 0);
			print_tt_token(pctx, NULL, ")");
			if (node->children->next->children == NULL)
				print_tt_token(pctx, NULL, ";");
			else
				c_print_node(pctx, node->children->next->children, 0);

			if (node->children->next->next != NULL) {
				print_tt_token(pctx, NULL, "else");
				if (node->children->next->next->children == NULL)
					print_tt_token(pctx, NULL, ";");
				else
					c_print_node(pctx, node->children->next->next->children, 0);
			}
			break;
		case 's': /* switch */
			print_tt_token(pctx, node, NULL);
			print_tt_token(pctx, NULL, "(");
			c_print_node(pctx, node->children, 0);
			print_tt_token(pctx, NULL, ")");
			c_print_node(pctx, node->children->next, 0);
			break;
		case 'b': /* break; */
		case 'c': /* continue */
			print_tt_token(pctx, node, NULL);
			print_tt_token(pctx, NULL, ";");
			break;
		case 'g': /* goto */
		case 'r': /* return */
			print_tt_token(pctx, node, NULL);
			c_print_node(pctx, node->children, 0);
			print_tt_token(pctx, NULL, ";");
			break;
		case ':': /* asm statements */
		case ',':
			print_tt_token(pctx, NULL, node->name);
			break;
		case '\0':
			/* empty statement */
			print_tt_token(pctx, node, ";");
			break;
		default:
			fprintf(stderr, "Unknown statement type %s\n", node->name);
			abort();
	}
}

static void c_print_attr_(c_print_t *pctx, const c99c_node_t *anode, int prec)
{
	for(;anode != NULL; anode = anode->next) {
		if (anode->type != T_ATTR) {
			fprintf(stderr, "unexpected node in attr list\n");
			abort();
		}
		print_tt_token(pctx, anode, anode->name);
		c_print_nodes(pctx, anode->children, ", ", prec);
	}
}


static void c_print_attr(c_print_t *pctx, const c99c_node_t *n, int prec)
{
	print_tt_token(pctx, n, "__attribute__");
	print_tt_token(pctx, NULL, "(");
	print_tt_token(pctx, NULL, "(");
	c_print_attr_(pctx, n->children, prec);
	print_tt_token(pctx, NULL, ")");
	print_tt_token(pctx, NULL, ")");
}

static void c_print_attrs(c_print_t *pctx, const c99c_node_t *parent_node, int allow_pre, int prec)
{
	const c99c_node_t *n;

	for(n = parent_node->children; n != NULL; n = n->next) {
		if (n->type == T_ATTRLIST) {
			if (allow_pre || attrlist_is_post(n)) {
				c_print_attr(pctx, n, prec);
			}
		}
	}
}

static void c_print_pre_attrs(c_print_t *pctx, const c99c_node_t *parent_node, int prec)
{
	const c99c_node_t *n;

	for(n = parent_node->children; n != NULL; n = n->next) {
		if (n->type == T_ATTRLIST) {
			if (attrlist_is_pre(n)) {
				c_print_attr(pctx, n, prec);
			}
		}
	}
}



/* if child prec <= parent prec then add parenthesis: (child) */
static void c_print_node(c_print_t *pctx, const c99c_node_t *node, int pprec)
{
	int prec;

	if (node == NULL)
		return;

	prec = getprec(pctx, node);

	switch(node->type) {
		case T_STATEMENT:
			c_print_statement(pctx, node);
			break;
		case T_OP:
			/* workaround for -(-i) vs --i, 16 == unary op */
			if (prec < pprec || (prec == 16 && pprec == 16 && (*node->name == '-' || *node->name == '+'))) {
/*				print_tt_token(pctx, NULL, "("); TODO #1 don't add */
				c_print_op(pctx, node, prec);
/*				print_tt_token(pctx, NULL, ")"); TODO #1 don't add */
			}
			else
				c_print_op(pctx, node, prec);
			break;
		case T_PRAGMA:
			print_tt_token(pctx, node, NULL);
/* TODO: without token tree:			printf("\n#pragma %s\n", node->name);*/
			break;
		case T_ICONST:
		case T_FCONST:
			print_tt_token(pctx, node, node->name);
			if (node->children != NULL) {
				fprintf(stderr, "ERROR: unexpected child node of %s:%s\n", c99c_node_type_names[node->type], node->name);
				exit(1);
			}
			break;

		case T_SCONST:
			print_tt_token(pctx, node, NULL);
			if (node->children != NULL) { /* for asm */
				print_tt_token(pctx, NULL, "(");
				c_print_nodes(pctx, node->children, ", ", 2);
				print_tt_token(pctx, NULL, ")");
			}
			break;

		case T_QUALIFIER:
		case T_CLASS:
		case T_TYPESPEC:
			if (node->children != NULL) {
				fprintf(stderr, "ERROR: unexpected child node of %s:%s\n", c99c_node_type_names[node->type], node->name);
				exit(1);
			}
		case T_TYPECUSTOM:
			print_tt_token(pctx, node, NULL);
			break;
		case T_TYPEOF:
			print_tt_token(pctx, node, "typeof");
			print_tt_token(pctx, NULL, "(");
			c_print_node(pctx, node->children, prec);
			print_tt_token(pctx, NULL, ")");
			break;
		case T_TYPE:
		case T_FUNCDEF:
		case T_CONDITION:
			if (prec < pprec)
				print_tt_eat(pctx, NULL, "(");
			c_print_nodes(pctx, node->children, NULL, prec);
			if (prec < pprec)
				print_tt_eat(pctx, NULL, ")");
			break;
		case T_EXPRBRACKET:
			print_tt_token(pctx, node, "(");
			c_print_nodes(pctx, node->children, ", ", prec+1);
			print_tt_token(pctx, NULL, ")");
			break;
		case T_EXPRLIST:
			if (prec < pprec)
				print_tt_eat(pctx, NULL, "(");
			c_print_nodes(pctx, node->children, ", ", prec+1);
			if (prec < pprec)
				print_tt_eat(pctx, NULL, ")");
			break;
		case T_SCONSTLIST:
			c_print_nodes(pctx, node->children, " ", prec);
			break;
		case T_TYPEDEF: /* same as decl */
			print_tt_token(pctx, node, "typedef");

		case T_ARG: /* either decl or expr, this works for both */
		case T_DECL:
			if (node->children == NULL)
				break;
			c_print_node(pctx, node->children, prec);

			print_tt_eat(pctx, NULL, WHITESPACE);
			c_print_nodes(pctx, node->children->next, ", ", prec);
			if (node->type != T_ARG)
				print_tt_eat(pctx, NULL, ";");
			break;
		case T_STRUCT:
			print_tt_token(pctx, NULL, "struct");
			c_print_pre_attrs(pctx, node, prec);
			print_tt_token(pctx, node, node->name == NULL ? "" : node->name);
			goto struni;
		case T_UNION:
			print_tt_token(pctx, NULL, "union");
			print_tt_token(pctx, node, node->name == NULL ? "" : node->name);
			struni:;
			/* empty struct{}, union{} always have a dummy decl child */
			print_tt_eat(pctx, NULL, "{");
			c_print_nodes(pctx, node->children, NULL, prec);
			print_tt_eat(pctx, NULL, "}");
			break;
		case T_ENUM:
			print_tt_token(pctx, NULL, "enum");
			print_tt_token(pctx, node, node->name == NULL ? "" : node->name);
			/* TODO: empty enum{}; */
			if (node->children) {
				print_tt_token(pctx, NULL, "{");
				c_print_nodes(pctx, node->children, ",\n", prec);
				print_tt_token(pctx, NULL, "}");
			}
			break;
		case T_INIT:
			print_tt_token(pctx, node, "=");
			c_print_nodes(pctx, node->children, NULL, prec);
			break;
		case T_TYPENAME:
			print_tt_id(pctx, node);
			pend_space(pctx);
			break;
		case T_ID:
			c_print_id(pctx, node);
			break;
		case T_ARGLIST:
			if (node->name && strcmp(node->name, "K&R/2") == 0) {
				printf("\n");
				c_print_nodes(pctx, node->children, ";", 0);
				if (node->children) {
					print_tt_token(pctx, NULL, ";");
					printf("\n");
				}
			}
			else {
				print_tt_token(pctx, node, "(");
				c_print_nodes(pctx, node->children, ", ", 0);
				if (node->loc.twin_parent != NULL) /* arglist inserted into the tree without token tree */
					print_tt_token(pctx, NULL, ")");
				else
					printf(")");
			}
			break;

		case T_CSE:
			print_tt_token(pctx, node, "(");
			print_tt_eat(pctx, NULL, "{");
			c_print_nodes(pctx, node->children, "\n", prec);
			if (node->children)
				print_tt_token(pctx, NULL, NEWLINE);
			print_tt_eat(pctx, NULL, "}");
			print_tt_token(pctx, NULL, ")");
			break;
		case T_BLOCK:
			if (prec < pprec)
				print_tt_eat(pctx, node, "(");
			if (!node->name || strcmp(node->name, "virtual") != 0)
				print_tt_token(pctx, node, "{");
			c_print_nodes(pctx, node->children, "\n", prec);
			if (node->children)
				print_tt_token(pctx, NULL, NEWLINE);
			if (!node->name || strcmp(node->name, "virtual") != 0)
				print_tt_eat(pctx, NULL, "}");
			if (prec < pprec)
				print_tt_eat(pctx, node, ")");
			break;
		case T_INITLIST:
			print_tt_token(pctx, node, "{");
			c_print_nodes(pctx, node->children, ", ", prec);
			print_tt_token(pctx, NULL, "}");
			break;
		case T_ASM:
			print_tt_token(pctx, node, "asm");
			print_tt_eat(pctx, NULL, "(");
			c_print_nodes(pctx, node->children, ",\n", prec);
			print_tt_eat(pctx, NULL, ")");
			break;
		case T_CALL:
			c_print_nodes(pctx, node->children, NULL, prec);
			break;
		case T_LABEL:
			switch(*node->name) {
				case 'd': /* default */
					print_tt_token(pctx, node, "default");
					print_tt_token(pctx, NULL, ":");
					break;
				case 'c': /* case */
					print_tt_token(pctx, node, "case");
					print_tt_token(pctx, NULL, " ");
					c_print_node(pctx, node->children, prec);
					print_tt_token(pctx, NULL, ":");
					break;
				case 'g': /* goto label*/
					c_print_node(pctx, node->children, prec);
					print_tt_token(pctx, NULL, ":");
					break;
				default:
					fprintf(stderr, "ERROR: invalid label type %s\n", node->name);
					exit(1);
					break;
			}
			if (node->next == NULL || node->next->type == T_DECL)
				print_tt_token(pctx, NULL, ";");
			break;
		case T_BITS:
			print_tt_token(pctx, node, ":");
			c_print_node(pctx, node->children, prec);
			break;
		case T_ELLIPSIS:
			print_tt_token(pctx, node, "...");
			break;
		case T_THEN:
		case T_ELSE:
			fprintf(stderr, "Internal error: unexpected token %s:%s\n", c99c_node_type_names[node->type], node->name);
			abort();

		case T_ATTRLIST:
			if (node->name == NULL) {
				print_tt_token(pctx, node, "__attribute__");
				print_tt_token(pctx, NULL, "(");
				print_tt_token(pctx, NULL, "(");
				c_print_nodes(pctx, node->children, ", ", prec);
				print_tt_token(pctx, NULL, ")");
				print_tt_token(pctx, NULL, ")");
			} /* else it is pre/post which is printed with IDs */
			break;
		case T_ATTR:
			print_tt_token(pctx, node, node->name);
			c_print_nodes(pctx, node->children, ", ", prec);
			break;

		case T_INVALID:
		case T_DEBUG:
			/* ignore these */
			break;
		default:
			printf("`%s:%s` ", c99c_node_type_names[node->type], node->name);
	}
}

static void c_print_nodes(c_print_t *pctx, const c99c_node_t *node, const char *sep_, int prec)
{
	int first;
	char *end, *sep;
	const char *ws = NULL;

	if (sep_ != NULL) {
		sep = c99c_strclone(sep_);
		end = sep + strlen(sep) - 1;
		if (*end == ' ') {
			*end = '\0';
			ws = WHITESPACE;
		}
		else if (*end == '\n') {
			*end = '\0';
			ws = NEWLINE;
		}
	}
	else
		sep = NULL;

	for(first = 1; node != NULL; node = node->next, first = 0) {
		if (!first && (sep != NULL)) {
			if (*sep != '\0') {
				if (node->loc.is_local)
					print_tt_ws_upto(pctx, sep);
				else
					print_tt_token(pctx, NULL, sep);
			}
		}

		if (first || node->loc.is_local)
			c_print_node(pctx, node, prec);
	}

	if (sep != NULL)
		free(sep);
}

int C99Nprint_c(c99tree_cc_t *ctx, void **nprint_ctx, const char *prefix, const c99c_node_t *node, print_state_t state)
{
	c_print_t *pctx;

	switch(state) {
		case PS_HEAD:
			pctx = malloc(sizeof(c_print_t));

			if (node->type == T_GRAMMAR_TREE) {
				if (node->next != NULL) {
					pctx->next_token = node->next->children;
					pctx->auto_ws = 0;
				}
				else {
					pctx->next_token = NULL;
					pctx->auto_ws = 1;
				}
			}
			else if ((node->parent->type == T_GRAMMAR_TREE) && (node->parent->children == node))
				pctx->next_token = node->parent->next->children; /* node is the first node in the grammar tree - print full file */
			else
				pctx->next_token = node->loc.twin_parent; /* print only from the given node */

			pctx->ctx = ctx;
			*nprint_ctx = pctx;
			c_print_nodes(pctx, node->children, "\n", 0);
			print_tt_rest(pctx);

/*			printf("\n"); .c must end with \n */

			free(pctx);
			break;
		case PS_OPENING:
		case PS_CLOSING:
			break;
		case PS_TAIL:
			break;
	}
	return 1; /* stop central processing after the root element */
}


